from typing import Dict, Optional, Union
from uuid import UUID

import requests
from pydantic import BaseModel, HttpUrl, validate_model, ValidationError, validator
from requests import RequestException

from basemodels.constants import SUPPORTED_CONTENT_TYPES


# New type
class AtLeastTenCharUrl(HttpUrl):
    min_length = 10


class TaskDataEntry(BaseModel):
    """
    Taskdata file format:

    [
      {
        "task_key": "407fdd93-687a-46bb-b578-89eb96b4109d",
        "datapoint_uri": "https://domain.com/file1.jpg",
        "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa"
      },
      {
        "task_key": "20bd4f3e-4518-4602-b67a-1d8dfabcce0c",
        "datapoint_uri": "https://domain.com/file2.jpg",
        "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa"
      }
    ]
    """

    task_key: Optional[UUID]
    datapoint_uri: HttpUrl

    @validator("datapoint_uri", always=True)
    def validate_datapoint_uri(cls, value):
        if len(value) < 10:
            raise ValidationError("datapoint_uri need to be at least 10 char length.")

    @validator("metadata")
    def validate_metadata(cls, value):
        if value is None:
            return value

        if len(value) > 10:
            raise ValidationError("10 key max. in metadata")

        if len(str(value)) > 1024:
            raise ValidationError("metadata should be < 1024")

        return value

    datapoint_hash: Optional[str]
    metadata: Optional[Dict[str, Optional[Union[str, int, float]]]]


def validate_content_type(uri: str) -> None:
    """Validate uri content type"""
    try:
        response = requests.head(uri)
        response.raise_for_status()
    except RequestException as e:
        raise ValidationError(
            f"taskdata content type ({uri}) validation failed", TaskDataEntry()
        ) from e

    content_type = response.headers.get("Content-Type", "")
    if content_type not in SUPPORTED_CONTENT_TYPES:
        raise ValidationError(
            f"taskdata entry datapoint_uri has unsupported type {content_type}",
            TaskDataEntry(),
        )


def validate_taskdata_entry(value: dict, validate_image_content_type: bool) -> None:
    """Validate taskdata entry"""
    if not isinstance(value, dict):
        raise ValidationError("taskdata entry should be dict", TaskDataEntry())

    *_, validation_error = validate_model(TaskDataEntry, value)
    if validation_error:
        raise validation_error

    if validate_image_content_type:
        validate_content_type(value["datapoint_uri"])



import unittest

from pydantic.error_wrappers import ValidationError
from basemodels import Preprocess, Pipeline


class PipelineTest(unittest.TestCase):
    def test_preprocess(self):
        config = {}
        p = Preprocess(pipeline=Pipeline.FaceBlurPipeline, config=config)

        self.assertEqual(p.pipeline, Pipeline.FaceBlurPipeline)
        self.assertEqual(p.config, config)

        p = Preprocess(pipeline=Pipeline.FaceBlurPipeline)

        self.assertIsNone(p.config)

    def test_preprocess_raise(self):
        with self.assertRaises(ValidationError):
            Preprocess()

        with self.assertRaises(ValidationError):
            Preprocess(pipeline="")

        with self.assertRaises(ValidationError):
            Preprocess(pipeline=Pipeline.FaceBlurPipeline, config=1)

    def test_preprocess_to_dict(self):
        config = {"radius": 3}
        p = Preprocess(pipeline=Pipeline.FaceBlurPipeline, config=config)

        self.assertEqual(
            p.to_dict(), {"pipeline": Pipeline.FaceBlurPipeline.value, "config": config}
        )

        p = Preprocess(pipeline=Pipeline.FaceBlurPipeline)

        self.assertEqual(p.to_dict(), {"pipeline": Pipeline.FaceBlurPipeline.value})



from enum import Enum
from uuid import UUID
from typing import Optional, List, Dict, Any
from pydantic import (
    BaseModel,
    root_validator,
    ValidationError,
    conint,
    confloat,
    validator,
)


class RestrictedAudienceBrowserEnum(str, Enum):
    mobile = "mobile"
    tablet = "tablet"
    desktop = "desktop"
    modern_browser = "modern_browser"


class RestrictedAudienceConfidenceEnum(str, Enum):
    minimum_client_confidence = "minimum_client_confidence"


class RestrictedAudienceScore(BaseModel):
    score: confloat(ge=0, le=1)


class RestrictedAudience(BaseModel):
    lang: Optional[List[Dict[str, RestrictedAudienceScore]]]
    country: Optional[List[Dict[str, RestrictedAudienceScore]]]
    sitekey: Optional[List[Dict[str, RestrictedAudienceScore]]]
    serverdomain: Optional[List[Dict[str, RestrictedAudienceScore]]]
    browser: Optional[
        List[Dict[RestrictedAudienceBrowserEnum, RestrictedAudienceScore]]
    ]
    confidence: Optional[
        List[Dict[RestrictedAudienceConfidenceEnum, RestrictedAudienceScore]]
    ]
    reason: Optional[List[Dict[str, RestrictedAudienceScore]]]

    min_difficulty: Optional[conint(ge=0, le=4, strict=True)]
    min_user_score: Optional[confloat(ge=0, le=1)]
    max_user_score: Optional[confloat(ge=0, le=1)]

    launch_group_id: Optional[conint(ge=0, strict=True)]

    interests: Optional[List[conint(strict=True)]]

    def dict(self, **kwargs):
        kwargs["exclude_unset"] = True
        return super().dict(**kwargs)

    def json(self, **kwargs):
        kwargs["exclude_unset"] = True
        return super().json(**kwargs)

    @root_validator()
    def validate_score_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        for entry, value in values.items():
            if value is None:
                continue
            if entry in [
                "lang",
                "country",
                "browser",
                "sitekey",
                "serverdomain",
                "confidence",
            ]:
                if isinstance(value, list):
                    for restriction in value:
                        if len(restriction) > 1:
                            raise ValueError("only 1 element per list item is allowed")
                        key = next(iter(restriction))
                        if entry in ["lang", "country", "sitekey"]:
                            if str(key) != str(key).lower():
                                raise ValueError("use lowercase")
        return values

    @validator("sitekey")
    def validate_sitekey(cls, value):
        if value is not None:
            for restriction in value:
                for sitekey in restriction:
                    try:
                        UUID(sitekey)
                    except:
                        raise ValidationError("invalid sitekey")
        return value



import uuid
import requests
from requests.exceptions import RequestException
from typing_extensions import Literal
from typing import Dict, Union, List, Optional
from enum import Enum
from uuid import UUID, uuid4
from .data.groundtruth import validate_groundtruth_entry
from .data.taskdata import validate_taskdata_entry
from pydantic import (
    BaseModel,
    validator,
    ValidationError,
    validate_model,
    HttpUrl,
    AnyHttpUrl,
)
from pydantic.fields import Field
from decimal import Decimal
from .restricted_audience import RestrictedAudience
from basemodels.constants import JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION


# Validator function for taskdata and taskdata_uri fields
def validator_taskdata_uri(cls, value, values, **kwargs):
    taskdata = values.get("taskdata")
    taskdata_uri = values.get("taskdata_uri")
    if taskdata is not None and len(taskdata) > 0 and taskdata_uri is not None:
        raise ValidationError(
            "Specify only one of taskdata {} or taskdata_uri {}".format(
                taskdata, taskdata_uri
            )
        )
    return value


# A validator function for UUID fields
def validate_uuid(cls, value):
    return value or uuid.uuid4()


# Base job types
class BaseJobTypesEnum(str, Enum):
    image_label_binary = "image_label_binary"
    image_label_multiple_choice = "image_label_multiple_choice"
    text_free_entry = "text_free_entry"
    text_multiple_choice_one_option = "text_multiple_choice_one_option"
    text_multiple_choice_multiple_options = "text_multiple_choice_multiple_options"
    image_label_area_adjust = "image_label_area_adjust"
    image_label_area_select = "image_label_area_select"
    image_label_single_polygon = "image_label_single_polygon"
    image_label_multiple_polygons = "image_label_multiple_polygons"
    image_label_semantic_segmentation_one_option = (
        "image_label_semantic_segmentation_one_option"
    )
    image_label_semantic_segmentation_multiple_options = (
        "image_label_semantic_segmentation_multiple_options"
    )
    image_label_text = "image_label_text"
    multi_challenge = "multi_challenge"


# Return a request type validator function
class RequestTypeValidator(object):
    def __init__(self, multi_challenge: bool = True):
        self.multi_challenge = multi_challenge

    def validate(self, cls, value, values):
        """
        validate request types for all types of challenges
        multi_challenge should always have multi_challenge_manifests
        """
        if value == BaseJobTypesEnum.multi_challenge:
            if not self.multi_challenge:
                raise ValidationError("multi_challenge request is not allowed here.")
            if "multi_challenge_manifests" not in values:
                raise ValidationError(
                    "multi_challenge requires multi_challenge_manifests."
                )
        elif value in [
            BaseJobTypesEnum.image_label_multiple_choice,
            BaseJobTypesEnum.image_label_area_select,
        ]:
            if values.get("multiple_choice_min_choices", 1) > values.get(
                "multiple_choice_max_choices", 1
            ):
                raise ValidationError(
                    "multiple_choice_min_choices cannot be greater than multiple_choice_max_choices"
                )
        return value


# Shape types enum
class ShapeTypes(str, Enum):
    point = "point"
    bounding_box = "bounding_box"
    polygon = "polygon"


class Model(BaseModel):
    def to_primitive(self):
        return self.dict()

    # Helper function for using in the unittest
    def check(self, return_new=False):
        self.__class__.validate(self)
        out_dict, _, validation_error = validate_model(self.__class__, self.__dict__)
        if validation_error:
            raise validation_error

        # For compatibility with tests
        if return_new:
            return self.__class__(**out_dict)


class Webhook(Model):
    """Model for webhook configuration"""

    webhook_id: UUID
    chunk_completed: Optional[List[str]]
    job_completed: Optional[List[str]]

    # States that might be interesting later
    # job_skipped: List[str] = None
    # job_inserted : List[str] = None
    # job_activated : List[str] = None


class TaskData(BaseModel):
    """Objects within taskdata list in Manifest"""

    task_key: UUID
    datapoint_uri: AnyHttpUrl
    datapoint_hash: str = Field(..., min_length=10, strip_whitespace=True)


class RequestConfig(Model):
    """Definition of the request_config object in manifest"""

    version: int = 0
    shape_type: Optional[ShapeTypes]
    min_points: Optional[int]
    max_points: Optional[int]
    min_shapes_per_image: Optional[int]
    max_shapes_per_image: Optional[int]
    restrict_to_coords: Optional[bool]
    minimum_selection_area_per_shape: Optional[int]
    multiple_choice_max_choices: Optional[int] = 1
    multiple_choice_min_choices: Optional[int] = 1
    overlap_threshold: Optional[float]


class InternalConfig(Model):
    """Discarded from incoming manifests"""

    exchange: Optional[Dict[str, Union[str, int, float]]]
    reco: Optional[Dict[str, Union[str, int, float]]]
    repo: Optional[Dict[str, Union[str, int, float]]]
    other: Optional[Dict[str, Union[str, int, float]]]
    # Accept one layer of nested
    mitl: Optional[Dict[str, Union[str, int, float, Dict[str, Union[str, int, float]]]]]

    class Config:
        arbitrary_types_allowed = True


class NestedManifest(Model):
    """The nested manifest description for multi_challenge jobs"""

    # We will set a default dynamic value for job_id
    job_id: Optional[UUID]
    validate_job_id = validator("job_id", always=True, allow_reuse=True)(validate_uuid)

    request_type: BaseJobTypesEnum
    validate_request_type = validator("request_type", allow_reuse=True, always=True)(
        RequestTypeValidator(multi_challenge=False).validate
    )
    requester_restricted_answer_set: Optional[Dict[str, Dict[str, str]]]

    @validator("requester_restricted_answer_set", always=True)
    def validate_requester_restricted_answer_set(cls, value, values, **kwargs):
        """image_label_area_select should always have a single RAS set"""

        # validation runs before other params, so need to handle missing case
        if "request_type" not in values:
            raise ValidationError("request_type missing")
        if values["request_type"] == BaseJobTypesEnum.image_label_area_select:
            if not value or len(value.keys()) == 0:
                value = {"label": {}}
                values["requester_restricted_answer_set"] = value
        return value

    requester_description: Optional[str]
    requester_max_repeats: int = 100
    requester_min_repeats: int = 1
    requester_question: Optional[Dict[str, str]]
    requester_question_example: Optional[Union[HttpUrl, List[HttpUrl]]]

    @validator("requester_question_example")
    def validate_requester_question_example(cls, value, values, **kwargs):
        # validation runs before other params, so need to handle missing case
        if not ("request_type" in values):
            raise ValidationError("request_type missing")

        # based on https://github.com/hCaptcha/hmt-basemodels/issues/27#issuecomment-590706643
        supports_lists = [
            BaseJobTypesEnum.image_label_area_select,
            BaseJobTypesEnum.image_label_binary,
        ]

        if isinstance(value, list) and not values["request_type"] in supports_lists:
            raise ValidationError("Lists are not allowed in this challenge type")
        return value

    unsafe_content: bool = False
    requester_accuracy_target: float = 0.1

    request_config: Optional[RequestConfig]

    # Groundtruth data is stored as a URL or optionally as an inlined json-serialized stringtype
    groundtruth_uri: Optional[HttpUrl]
    groundtruth: Optional[str]

    @validator("groundtruth", always=True)
    def validate_groundtruth(cls, v, values, **kwargs):
        if "groundtruth_uri" in values and "groundtruth" in values:
            raise ValidationError(
                "Specify only groundtruth_uri or groundtruth, not both."
            )
        return v

    # Configuration id -- XXX LEGACY
    confcalc_configuration_id: Optional[str]
    webhook: Optional[Webhook]

    class Config:
        arbitrary_types_allowed = True


class Manifest(Model):
    """The manifest description."""

    job_mode: Literal["batch", "online", "instant_delivery"]

    # We will set a default dynamic value for job_api_key
    job_api_key: Optional[UUID]

    # We will set a default dynamic value for job_id
    job_id: UUID = Field(default_factory=uuid4)

    job_total_tasks: int
    multi_challenge_manifests: Optional[List[NestedManifest]]
    request_type: BaseJobTypesEnum
    network: Optional[str]
    only_sign_results: bool = False
    public_results: bool = False

    requester_restricted_answer_set: Optional[Dict[str, Dict[str, str]]]

    requester_description: Optional[str]
    requester_max_repeats: int = 100
    requester_min_repeats: int = 1
    requester_question: Optional[Dict[str, str]]

    requester_question_example: Optional[Union[HttpUrl, List[HttpUrl]]]

    unsafe_content: bool = False
    task_bid_price: float
    oracle_stake: Decimal
    expiration_date: Optional[int]
    requester_accuracy_target: float = 0.1
    manifest_smart_bounty_addr: Optional[str]
    hmtoken_addr: Optional[str]
    minimum_trust_server: float = 0.1
    minimum_trust_client: float = 0.1
    recording_oracle_addr: Optional[str]
    reputation_oracle_addr: Optional[str]
    reputation_agent_addr: Optional[str]
    requester_pgp_public_key: Optional[str]
    ro_uri: Optional[str]
    repo_uri: Optional[str]
    batch_result_delivery_webhook: Optional[AnyHttpUrl]
    online_result_delivery_webhook: Optional[AnyHttpUrl]
    instant_result_delivery_webhook: Optional[AnyHttpUrl]

    request_config: Optional[RequestConfig]

    # If taskdata is directly provided
    taskdata: Optional[List[TaskData]]

    # If taskdata is separately stored
    taskdata_uri: Optional[AnyHttpUrl]

    # Groundtruth data is stored as a URL or optionally as an inlined json-serialized stringtype
    groundtruth_uri: Optional[AnyHttpUrl]
    groundtruth: Optional[str]

    # internal config options for param tests etc.
    internal_config: Optional[InternalConfig]

    # Configuration id
    confcalc_configuration_id: Optional[str]
    restricted_audience: Optional[RestrictedAudience] = {}

    webhook: Optional[Webhook]

    rejected_uri: Optional[AnyHttpUrl]
    rejected_count: Optional[int]

    # Validators

    @validator("requester_min_repeats")
    def validate_min_repeats(cls, v, values):
        """min repeats are required to be at least 4 if ilmc"""
        if values["request_type"] == "image_label_multiple_choice":
            return max(v, 4)
        return v

    @validator("groundtruth", always=True)
    def validate_groundtruth(cls, value, values):
        if "groundtruth_uri" in values and "groundtruth" in values:
            raise ValueError("Specify only groundtruth_uri or groundtruth, not both.")
        return value

    @validator("requester_restricted_answer_set", always=True)
    def validate_requester_restricted_answer_set(cls, value, values, **kwargs):
        """image_label_area_select should always have a single RAS set"""
        # validation runs before other params, so need to handle missing case
        if not ("request_type" in values):
            raise ValueError("request_type missing")

        if values["request_type"] == BaseJobTypesEnum.image_label_area_select:
            if not value or len(value.keys()) == 0:
                value = {"label": {}}
                values["requester_restricted_answer_set"] = value
        return value

    @validator("requester_question_example")
    def validate_requester_question_example(cls, value, values, **kwargs):
        # validation runs before other params, so need to handle missing case
        if not ("request_type" in values):
            raise ValueError("request_type missing")

        # based on https://github.com/hCaptcha/hmt-basemodels/issues/27#issuecomment-590706643
        supports_lists = [
            BaseJobTypesEnum.image_label_area_select,
            BaseJobTypesEnum.image_label_binary,
        ]

        if isinstance(value, list) and not (values["request_type"] in supports_lists):
            raise ValueError("Lists are not allowed in this challenge type")
        return value

    validate_taskdata_uri = validator("taskdata_uri", allow_reuse=True, always=True)(
        validator_taskdata_uri
    )
    validate_taskdata = validator("taskdata", allow_reuse=True, always=True)(
        validator_taskdata_uri
    )
    validate_request_type = validator("request_type", allow_reuse=True, always=True)(
        RequestTypeValidator().validate
    )
    validate_job_api_key = validator("job_api_key", always=True, allow_reuse=True)(
        validate_uuid
    )
    validate_job_id = validator("job_id", always=True, allow_reuse=True)(validate_uuid)

    class Config:
        arbitrary_types_allowed = True


def validate_groundtruth_uri(manifest: dict):
    """
    Validate groundtruth_uri
    Returns entries count if succeeded
    """
    request_type = manifest.get("request_type", "")
    validate_image_content_type = request_type in JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION
    uri_key = "groundtruth_uri"
    uri = manifest.get(uri_key)
    if uri is None:
        return
    try:
        response = requests.get(uri, timeout=1)
        response.raise_for_status()

        entries_count = 0
        data = response.json()
        if isinstance(data, dict):
            for k, v in data.items():
                entries_count += 1
                validate_groundtruth_entry(
                    k, v, request_type, validate_image_content_type
                )
                validate_image_content_type = False
        else:
            for v in data:
                entries_count += 1
                validate_groundtruth_entry(
                    "", v, request_type, validate_image_content_type
                )
                validate_image_content_type = False

    except (ValidationError, RequestException) as e:
        raise ValidationError(f"{uri_key} validation failed: {e}", Manifest) from e

    if entries_count == 0:
        raise ValidationError(f"fetched {uri_key} is empty", Manifest)


def validate_taskdata_uri(manifest: dict):
    """
    Validate taskdata_uri
    Returns entries count if succeeded
    """
    request_type = manifest.get("request_type", "")
    validate_image_content_type = request_type in JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION
    uri_key = "taskdata_uri"
    uri = manifest.get(uri_key)
    if uri is None:
        return
    try:
        response = requests.get(uri, timeout=1)
        response.raise_for_status()

        entries_count = 0
        data = response.json()
        for v in data:
            entries_count += 1
            validate_taskdata_entry(v, validate_image_content_type)
            # We want to validate only first entry for content type
            validate_image_content_type = False

    except (ValidationError, RequestException) as e:
        raise ValidationError(f"{uri_key} validation failed: {e}", Manifest) from e

    if entries_count == 0:
        raise ValidationError(f"fetched {uri_key} is empty", Manifest)


def validate_manifest_uris(manifest: dict):
    """Fetch & validate manifest's remote objects"""
    validate_taskdata_uri(manifest)
    validate_groundtruth_uri(manifest)



from typing import List, Optional, Union

import requests
from pydantic import BaseModel, HttpUrl, ValidationError
from requests import RequestException
from typing_extensions import Literal

from basemodels.constants import SUPPORTED_CONTENT_TYPES


def create_wrapper_model(type):
    class WrapperModel(BaseModel):
        data: Optional[type]

        class Config:
            arbitrary_types_allowed = True

    return WrapperModel


def validate_wrapper_model(Model, data):
    Model.validate({"data": data})


groundtruth_entry_key_type = HttpUrl
GroundtruthEntryKeyModel = create_wrapper_model(groundtruth_entry_key_type)
"""
Groundtruth file format for `image_label_binary` job type:

{
  "https://domain.com/file1.jpeg": ["false", "false", "false"],
  "https://domain.com/file2.jpeg": ["true", "true", "true"]
}
"""
ilb_groundtruth_entry_type = List[Literal["true", "false"]]
ILBGroundtruthEntryModel = create_wrapper_model(ilb_groundtruth_entry_type)
"""
Groundtruth file format for `image_label_multiple_choice` job type:

{
  "https://domain.com/file1.jpeg": [
    ["cat"],
    ["cat"],
    ["cat"]
  ],
  "https://domain.com/file2.jpeg": [
    ["dog"],
    ["dog"],
    ["dog"]
  ]
}
"""
ilmc_groundtruth_entry_type = List[List[str]]
ILMCGroundtruthEntryModel = create_wrapper_model(ilmc_groundtruth_entry_type)


class ILASGroundtruthEntry(BaseModel):
    entity_name: float
    entity_type: str
    entity_coords: List[Union[int, float]]


"""
Groundtruth file format for `image_label_area_select` job type

{
  "https://domain.com/file1.jpeg": [
    [
      {
        "entity_name": 0,
        "entity_type": "gate",
        "entity_coords": [275, 184, 454, 183, 453, 366, 266, 367]
      }
    ]
  ]
}
"""
ilas_groundtruth_entry_type = List[List[ILASGroundtruthEntry]]
ILASGroundtruthEntryModel = create_wrapper_model(ilas_groundtruth_entry_type)

groundtruth_entry_models_map = {
    "image_label_binary": ILBGroundtruthEntryModel,
    "image_label_multiple_choice": ILMCGroundtruthEntryModel,
    "image_label_area_select": ILASGroundtruthEntryModel,
}


def validate_content_type(uri: str) -> None:
    """Validate uri content type"""
    try:
        response = requests.head(uri)
        response.raise_for_status()
    except RequestException as e:
        raise ValidationError(
            f"groundtruth content type ({uri}) validation failed",
            GroundtruthEntryKeyModel,
        ) from e

    content_type = response.headers.get("Content-Type", "")
    if content_type not in SUPPORTED_CONTENT_TYPES:
        raise ValidationError(
            f"groundtruth entry has unsupported type {content_type}",
            GroundtruthEntryKeyModel,
        )


def validate_groundtruth_entry(
    key: str,
    value: Union[dict, list],
    request_type: str,
    validate_image_content_type: bool,
):
    """Validate key & value of groundtruth entry based on request_type"""
    GroundtruthEntryValueModel = groundtruth_entry_models_map.get(request_type)

    if GroundtruthEntryValueModel is None:
        return

    validate_wrapper_model(GroundtruthEntryKeyModel, key)
    validate_wrapper_model(GroundtruthEntryValueModel, value)

    if validate_image_content_type:
        validate_content_type(key)



#!/usr/bin/env python3
from pydantic import ValidationError
from typing import Any

import basemodels
from uuid import uuid4
from copy import deepcopy

from basemodels.manifest.data.taskdata import TaskDataEntry

import unittest
import httpretty
import json

from basemodels.manifest.restricted_audience import RestrictedAudience

CALLBACK_URL = "http://google.com/webback"
FAKE_URL = "http://google.com/fake"
IMAGE_LABEL_BINARY = "image_label_binary"

REP_ORACLE = "0x61F9F0B31eacB420553da8BCC59DC617279731Ac"
REC_ORACLE = "0xD979105297fB0eee83F7433fC09279cb5B94fFC6"
FAKE_ORACLE = "0x1413862c2b7054cdbfdc181b83962cb0fc11fd92"


# A helper function for create manifest models based on model library
def create_manifest(data: dict):
    return basemodels.Manifest.construct(**data)


# A helper function for create nested manifest models based on model library
def create_nested_manifest(data: dict):
    return basemodels.NestedManifest.construct(**data)


# A helper function for create nested manifest models based on model library
def create_webhook(data: dict):
    return basemodels.Webhook.construct(**data)


# Json serializer for models based on libraries
def to_json(model):
    # Pydantic json serializer
    return model.json()


# A helper function for providing validatation function based on libraries
def validate_func(model):
    return model.check


def a_manifest(
    number_of_tasks=100,
    bid_amount=1.0,
    oracle_stake=0.05,
    expiration_date=0,
    minimum_trust=0.1,
    request_type=IMAGE_LABEL_BINARY,
    request_config=None,
    job_mode="batch",
    multi_challenge_manifests=None,
) -> Any:
    internal_config = {"exchange": {"a": 1, "b": "c"}}
    model = {
        "requester_restricted_answer_set": {
            "0": {"en": "English Answer 1"},
            "1": {
                "en": "English Answer 2",
                "answer_example_uri": "https://hcaptcha.com/example_answer2.jpg",
            },
        },
        "job_mode": job_mode,
        "request_type": request_type,
        "internal_config": internal_config,
        "multi_challenge_manifests": multi_challenge_manifests,
        "unsafe_content": False,
        "task_bid_price": bid_amount,
        "oracle_stake": oracle_stake,
        "expiration_date": expiration_date,
        "minimum_trust_server": minimum_trust,
        "minimum_trust_client": minimum_trust,
        "requester_accuracy_target": minimum_trust,
        "recording_oracle_addr": REC_ORACLE,
        "reputation_oracle_addr": REP_ORACLE,
        "reputation_agent_addr": REP_ORACLE,
        "instant_result_delivery_webhook": CALLBACK_URL,
        "requester_question": {"en": "How much money are we to make"},
        "requester_question_example": FAKE_URL,
        "job_total_tasks": number_of_tasks,
        "taskdata_uri": FAKE_URL,
    }

    if request_config:
        model.update({"request_config": request_config})

    manifest = create_manifest(model)
    validate_func(manifest)()
    return manifest


def a_nested_manifest(
    request_type=IMAGE_LABEL_BINARY, minimum_trust=0.1, request_config=None
) -> Any:
    model = {
        "requester_restricted_answer_set": {
            "0": {"en": "English Answer 1"},
            "1": {
                "en": "English Answer 2",
                "answer_example_uri": "https://hcaptcha.com/example_answer2.jpg",
            },
        },
        "request_type": request_type,
        "requester_accuracy_target": minimum_trust,
        "requester_question": {"en": "How much money are we to make"},
        "requester_question_example": FAKE_URL,
    }

    if request_config:
        model.update({"request_config": request_config})

    manifest = create_nested_manifest(model)
    validate_func(manifest)()

    return manifest


TASK = {
    "task_key": "407fdd93-687a-46bb-b578-89eb96b4109d",
    "datapoint_uri": "https://domain.com/file1.jpg",
    "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa",
    "metadata": {
        "key_1": "value_1",
        "key_2": "value_2",
    },
}


class ManifestTest(unittest.TestCase):
    """Manifest specific tests, validating that models work the way we want"""

    def test_basic_construction(self):
        """Tests that manifest can validate the test manifest properly."""
        a_manifest()

    def test_can_serialize(self):
        """validate that we can dump this to json in downstream services"""
        j = to_json(a_manifest())

    def test_can_fail_toconstruct(self):
        """Tests that the manifest raises an Error when called with falsy parameters."""
        a_manifest(-1)
        self.assertRaises(ValidationError, a_manifest, "invalid amount")

    def test_can_fail_toconstruct2(self):
        """Tests that validated fields can't be broken without an exception."""
        mani = a_manifest()
        mani.taskdata_uri = "test"
        self.assertRaises(ValidationError, validate_func(mani))

    def test_can_make_request_config_job(self):
        """Test that jobs with valid request_config parameter work"""
        manifest = a_manifest(
            request_type="image_label_area_select",
            request_config={"shape_type": "point", "overlap_threshold": 0.8},
        )

    def test_can_make_nested_request_config_job_single_nest(self):
        """Test that jobs with valid nested request_config parameter work"""
        nested_manifest = a_nested_manifest(
            request_type="image_label_area_select",
            request_config={"shape_type": "point"},
        )

        manifest = a_manifest(
            request_type="multi_challenge", multi_challenge_manifests=[nested_manifest]
        )

    def test_can_make_nested_request_config_job_multiple_nest(self):
        """Test that jobs with multiple valid nested request_config parameters work"""
        nested_manifest = a_nested_manifest(
            request_type="image_label_area_select",
            request_config={"shape_type": "point"},
        )

        nested_manifest_2 = a_nested_manifest(
            request_type="image_label_area_select",
            request_config={"shape_type": "point"},
        )

        manifest = a_manifest(
            request_type="multi_challenge",
            multi_challenge_manifests=[nested_manifest, nested_manifest_2],
        )

    def test_can_bad_request_config(self):
        """Test that an invalid shape_type in request_config will fail"""
        manifest = a_manifest()
        manifest.request_type = "image_label_area_select"
        manifest.request_config = {"shape_type": "not-a-real-option"}
        self.assertRaises(ValidationError, validate_func(manifest))

    def test_gets_default_restrictedanswerset(self):
        """Make sure that the image_label_area_select jobs get a default RAS"""
        model = {
            "job_mode": "batch",
            "request_type": "image_label_area_select",
            "unsafe_content": False,
            "task_bid_price": 1,
            "oracle_stake": 0.1,
            "expiration_date": 0,
            "minimum_trust_server": 0.1,
            "minimum_trust_client": 0.1,
            "requester_accuracy_target": 0.1,
            "recording_oracle_addr": REC_ORACLE,
            "reputation_oracle_addr": REP_ORACLE,
            "reputation_agent_addr": REP_ORACLE,
            "instant_result_delivery_webhook": CALLBACK_URL,
            "requester_question": {"en": "How much money are we to make"},
            "requester_question_example": FAKE_URL,
            "job_total_tasks": 5,
            "taskdata_uri": FAKE_URL,
        }

        manifest = create_manifest(model)

        func = validate_func(manifest)
        manifest = func(True)

        self.assertGreater(
            len(manifest.to_primitive()["requester_restricted_answer_set"].keys()), 0
        )

    def test_confcalc_configuration_id(self):
        """Test that key is in manifest"""
        manifest = a_manifest()
        manifest.confcalc_configuration_id = "test_conf_id"
        validate_func(manifest)()

        self.assertTrue("confcalc_configuration_id" in manifest.to_primitive())

    def test_url_or_list_for_example(self):
        """validates that we can supply a list or a url to example key"""
        model = a_manifest()

        model.requester_question_example = "https://test.com"
        self.assertTrue(validate_func(model)() is None)
        self.assertIsInstance(model.to_primitive()["requester_question_example"], str)

        model.requester_question_example = ["https://test.com"]
        self.assertTrue(validate_func(model)() is None)
        self.assertIsInstance(model.to_primitive()["requester_question_example"], list)

        model.requester_question_example = "non-url"
        self.assertRaises(ValidationError, validate_func(model))
        model.requester_question_example = ["non-url"]
        self.assertRaises(ValidationError, validate_func(model))

        # we now allow lists in non-ilb types
        model.request_type = "image_label_area_select"
        self.assertTrue(validate_func(model))

    def test_restricted_audience(self):
        """Test that restricted audience is in the Manifest"""
        manifest = a_manifest()

        restricted_audience = {
            "lang": [{"en-us": {"score": 0.9}}],
            "confidence": [{"minimum_client_confidence": {"score": 0.9}}],
            "min_difficulty": 2,
        }

        manifest.restricted_audience = RestrictedAudience(**restricted_audience)

        validate_func(manifest)()
        self.assertTrue("restricted_audience" in manifest.to_primitive())
        self.assertTrue(
            "minimum_client_confidence"
            in manifest.to_primitive()["restricted_audience"]["confidence"][0]
        )
        self.assertEqual(
            0.9,
            manifest.to_primitive()["restricted_audience"]["confidence"][0][
                "minimum_client_confidence"
            ]["score"],
        )
        self.assertTrue(
            "en-us" in manifest.to_primitive()["restricted_audience"]["lang"][0]
        )
        self.assertEqual(
            0.9,
            manifest.to_primitive()["restricted_audience"]["lang"][0]["en-us"]["score"],
        )
        self.assertEqual(
            2, manifest.to_primitive()["restricted_audience"]["min_difficulty"]
        )

    def test_parse_restricted_audience(self):
        """Test None fields are skipped in restricted audience"""
        restricted_audience = {"min_difficulty": 2}

        self.assertEqual(
            RestrictedAudience(**restricted_audience).dict(), {"min_difficulty": 2}
        )
        self.assertEqual(
            RestrictedAudience(**restricted_audience).json(), '{"min_difficulty": 2}'
        )

    def test_restricted_audience_only(self):
        def assert_raises(data):
            with self.assertRaises(ValidationError):
                RestrictedAudience(**data)

        for data in [
            {"lang": "us"},
            {"lang": [{"US": {"score": 1}}]},
            {"lang": [{"US": "US"}]},
            {"lang": [{"us": {"nonsense": 1}}]},
            {"lang": [{"us": {"score": -0.1}}]},
        ]:
            assert_raises(data)

        for data in [
            {"country": "us"},
            {"country": [{"US": {"score": 1}}]},
            {"country": [{"US": "US"}]},
            {"country": [{"us": {"nonsense": 1}}]},
            {"country": [{"us": {"score": -0.1}}]},
            {"country": [{"us": {"score": -0.1}}]},
        ]:
            assert_raises(data)

        for data in [
            {"browser": "desktop"},
            {"browser": [{"Desktop": {"score": 1}}]},
            {"browser": [{"desktop": "US"}]},
            {"browser": [{"desktop": {"nonsense": 1}}]},
            {"browser": [{"desktop": {"score": -0.1}}]},
        ]:
            assert_raises(data)

        sitekey = "9d98b147-dc5a-4ea4-82cf-0ced5b2434d2"
        for data in [
            {"sitekey": "sitekey"},
            {"sitekey": [{"9d98b147": {"score": 1}}]},
            {"sitekey": [{sitekey.upper(): {"score": 1}}]},
            {"sitekey": [{sitekey: 1}]},
            {"sitekey": [{sitekey: {"nonsense": 1}}]},
            {"sitekey": [{sitekey: {"score": -0.1}}]},
        ]:
            assert_raises(data)

        for data in [
            {"serverdomain": "serverdomain"},
            {"serverdomain": [{"serverdomain": 1}]},
            {"serverdomain": [{"serverdomain": {"nonsense": 1}}]},
            {"serverdomain": [{"serverdomain": {"score": -0.1}}]},
        ]:
            assert_raises(data)

        for data in [
            {"confidence": "confidence"},
            {"confidence": [{"MINIMUM_client_confidence": {"score": 1}}]},
            {"confidence": [{"minimum_client_confidence": 1}]},
            {"confidence": [{"minimum_client_confidence": {"nonsense": 1}}]},
            {"confidence": [{"minimum_client_confidence": {"score": -0.1}}]},
        ]:
            assert_raises(data)

        for data in [
            {"min_difficulty": "min_difficulty"},
            {"min_difficulty": -1},
            {"min_difficulty": 5},
            {"min_difficulty": 1.1},
        ]:
            assert_raises(data)

        for data in [
            {"min_user_score": "min_user_score"},
            {"min_user_score": -0.1},
            {"min_user_score": 1.1},
        ]:
            assert_raises(data)

        for data in [
            {"max_user_score": "max_user_score"},
            {"max_user_score": -0.1},
            {"max_user_score": 1.1},
        ]:
            assert_raises(data)

        for data in [
            {"launch_group_id": "launch_group_id"},
            {"launch_group_id": -3},
            {"launch_group_id": 1.1},
        ]:
            assert_raises(data)

        for data in [
            {"interests": 1},
            {"interests": {"mapped": 1}},
            {"interests": ["as", "string"]},
        ]:
            assert_raises(data)

        data = {
            "lang": [
                {"us": {"score": 0}},
                {"es": {"score": 0.5}},
                {"en-us": {"score": 1}},
            ],
            "country": [
                {"us": {"score": 0}},
                {"es": {"score": 0.5}},
                {"it": {"score": 1}},
            ],
            "browser": [
                {"tablet": {"score": 0.5}},
                {"desktop": {"score": 0}},
                {"mobile": {"score": 1}},
                {"modern_browser": {"score": 0.9}},
            ],
            "sitekey": [
                {str(uuid4()): {"score": 0.5}},
                {str(uuid4()): {"score": 0}},
                {str(uuid4()): {"score": 1}},
            ],
            "serverdomain": [
                {"1hcaptcha.com": {"score": 0.5}},
                {"2hcaptcha.com": {"score": 0}},
                {"3hcaptcha.com": {"score": 1}},
            ],
            "confidence": [{"minimum_client_confidence": {"score": 0.5}}],
            "min_difficulty": 2,
            "min_user_score": 0,
            "max_user_score": 0.3,
            "launch_group_id": 101,
            "interests": [1, 2, 3, 4],
        }

        RestrictedAudience(**data)

    def test_realistic_multi_challenge_example(self):
        """validates a realistic multi_challenge manifest"""
        obj = {
            "job_mode": "batch",
            "request_type": "image_label_area_select",
            "unsafe_content": False,
            "task_bid_price": 1,
            "oracle_stake": 0.1,
            "expiration_date": 0,
            "minimum_trust_server": 0.1,
            "minimum_trust_client": 0.1,
            "requester_accuracy_target": 0.1,
            "job_total_tasks": 1000,
            "recording_oracle_addr": REC_ORACLE,
            "reputation_oracle_addr": REP_ORACLE,
            "reputation_agent_addr": REP_ORACLE,
            "job_id": "c26c2e6a-41ab-4218-b39e-6314b760c45c",
            "request_type": "multi_challenge",
            "requester_question": {
                "en": "Please draw a bow around the text shown, select the best corresponding labels, and enter the word depicted by the image."
            },
            "multi_challenge_manifests": [
                {
                    "request_type": "image_label_area_select",
                    "job_id": "c26c2e6a-41ab-4218-b39e-6314b760c45c",
                    "requester_question": {
                        "en": "Please draw a bow around the text shown."
                    },
                    "request_config": {
                        "shape_type": "polygon",
                        "min_points": 1,
                        "max_points": 4,
                        "min_shapes_per_image": 1,
                        "max_shapes_per_image": 4,
                    },
                },
                {
                    "request_type": "image_label_multiple_choice",
                    "job_id": "c26c2e6a-41ab-4218-b39e-6314b760c45c",
                    "requester_question": {"en": "Select the corresponding label."},
                    "requester_restricted_answer_set": {
                        "print": {"en": "Print"},
                        "hand-writing": {"en": "Hand Writing"},
                    },
                    "request_config": {"multiple_choice_max_choices": 1},
                },
                {
                    "request_type": "image_label_multiple_choice",
                    "job_id": "c26c2e6a-41ab-4218-b39e-6314b760c45c",
                    "requester_question": {"en": "Select the corresponding labels."},
                    "requester_restricted_answer_set": {
                        "top-bottom": {"en": "Top to Bottom"},
                        "bottom-top": {"en": "Bottom to Top"},
                        "left-right": {"en": "Left to Right"},
                        "right-left": {"en": "Right to Left"},
                    },
                    "request_config": {"multiple_choice_max_choices": 1},
                },
                {
                    "request_type": "image_label_text",
                    "job_id": "c26c2e6a-41ab-4218-b39e-6314b760c45c",
                    "requester_question": {"en": "Please enter the word in the image."},
                },
            ],
            "taskdata": [
                {
                    "datapoint_hash": "sha1:5daf66c6031df7f8913bfa0b52e53e3bcd42aab3",
                    "datapoint_uri": "http://test.com/task.jpg",
                    "task_key": "2279daef-d10a-4b0f-85d1-0ccbf7c8906b",
                }
            ],
        }

        model = create_manifest(obj)
        # print(model.to_primitive())
        self.assertTrue(validate_func(model)() is None)

    def test_webhook(self):
        """Test that webhook is correct"""
        webhook = {
            "webhook_id": "c26c2e6a-41ab-4218-b39e-6314b760c45c",
            "job_completed": ["http://servicename:4000/api/webhook"],
        }

        webhook_model = create_webhook(webhook)
        validate_func(webhook_model)()
        self.assertTrue("webhook_id" in webhook_model.to_primitive())

        model = a_manifest()
        model.webhook = webhook
        validate_func(model)()
        self.assertTrue("webhook" in model.to_primitive())

    def test_invalid_example_images(self):
        """validate that only certain types of request_types can specify example images"""
        manifest = a_manifest().to_primitive()
        manifest["request_type"] = "image_label_multiple_choice"
        manifest["requester_question_example"] = ["a"]

        with self.assertRaises(ValidationError):
            validate_func(create_manifest(manifest))()

        del manifest["requester_question_example"]
        validate_func(create_manifest(manifest))()

    def test_default_only_sign_results(self):
        """Test whether flag 'only_sign_results' is False by default."""
        manifest = a_manifest()
        self.assertEqual(manifest.only_sign_results, False)

    def test_default_public_results(self):
        """Test whether flag 'public_results' is False by default."""
        manifest = a_manifest()
        self.assertEqual(manifest.public_results, False)


class ViaTest(unittest.TestCase):
    def test_via_legacy_case(self):
        """tests case with inner class_attributes"""
        content = {
            "datapoints": [
                {
                    "task_uri": "https://mydomain.com/image.jpg",
                    "metadata": {"filename": "image.jpg"},
                    "class_attributes": {
                        "0": {"class_attributes": {"dog": False, "cat": False}}
                    },
                    "regions": [
                        {
                            "region_attributes": {"region_key": "region_value"},
                            "shape_attributes": {
                                "coords": [1, 2, 3, 4, 5, 6, 7, 8.0],
                                "name": "shape_type",
                            },
                        }
                    ],
                }
            ]
        }

        # Also test the marshmallow model from the old package
        parsed = basemodels.ViaDataManifest(**content).dict()
        self.assertEqual(len(parsed["datapoints"]), 1)
        self.assertEqual(parsed["version"], 1)

    def test_via_v1_case(self):
        """tests case where we dont use the inner class_attributes"""
        content = {
            "datapoints": [
                {
                    "task_uri": "https://mydomain.com/image.jpg",
                    "metadata": {"filename": "image.jpg"},
                    "class_attributes": {"dog": False, "cat": False},
                    "regions": [
                        {
                            "region_attributes": {"region_key": "region_value"},
                            "shape_attributes": {
                                "coords": [1, 2, 3, 4, 5, 6, 7, 8.0],
                                "name": "shape_type",
                            },
                        }
                    ],
                }
            ]
        }

        # Also test the marshmallow model from the old package
        parsed = basemodels.ViaDataManifest(**content).dict()
        self.assertEqual(len(parsed["datapoints"]), 1)
        self.assertEqual(parsed["version"], 1)
        self.assertIn("dog", parsed["datapoints"][0]["class_attributes"])


@httpretty.activate
class TestValidateManifestUris(unittest.TestCase):
    def register_http_response(
        self,
        uri="https://uri.com",
        manifest=None,
        body=None,
        method=httpretty.GET,
        headers=None,
    ):
        headers = headers or {}
        httpretty.register_uri(method, uri, body=json.dumps(body), **headers)

    def validate_groundtruth_response(self, request_type, body):
        uri = "https://uri.com"
        manifest = {"groundtruth_uri": uri, "request_type": request_type}

        self.register_http_response(uri, manifest, body)

        basemodels.validate_manifest_uris(manifest)

    def test_no_uris(self):
        """should not raise if there are no uris to validate"""
        manifest = {}
        basemodels.validate_manifest_uris(manifest)

    def test_groundtruth_uri_ilb_valid(self):
        groundtruth_uri = "https://domain.com/123/file1.jpeg"
        body = {
            groundtruth_uri: ["false", "false", "false"],
            "https://domain.com/456/file2.jpeg": ["false", "true", "false"],
        }

        self.register_http_response(
            groundtruth_uri,
            method=httpretty.HEAD,
            headers={"Content-Type": "image/jpeg"},
        )
        self.validate_groundtruth_response("image_label_binary", body)

    def test_groundtruth_uri_ilb_invalid(self):
        body = {"not_uri": ["false", "false", True]}

        with self.assertRaises(ValidationError):
            self.validate_groundtruth_response("image_label_binary", body)

    def test_groundtruth_uri_ilb_invalid_format(self):
        """should raise if groundtruth_uri contains array instead of object"""
        body = [{"key": "value"}]

        with self.assertRaises(ValidationError):
            self.validate_groundtruth_response("image_label_binary", body)

    def test_groundtruth_uri_ilmc_valid(self):
        groundtruth_uri = "https://domain.com/file1.jpeg"
        body = {
            groundtruth_uri: [["cat"], ["cat"], ["cat"]],
            "https://domain.com/file2.jpeg": [["dog"], ["dog"], ["dog"]],
        }

        self.register_http_response(
            groundtruth_uri,
            method=httpretty.HEAD,
            headers={"Content-Type": "image/jpeg"},
        )
        self.validate_groundtruth_response("image_label_multiple_choice", body)

    def test_groundtruth_uri_ilmc_invalid_key(self):
        body = {"not_uri": [["cat"], ["cat"], ["cat"]]}

        with self.assertRaises(ValidationError):
            self.validate_groundtruth_response("image_label_multiple_choice", body)

    def test_groundtruth_uri_ilmc_invalid_value(self):
        body = {
            "https://domain.com/file1.jpeg": [True, False],
        }

        with self.assertRaises(ValidationError):
            self.validate_groundtruth_response("image_label_multiple_choice", body)

    def test_groundtruth_uri_ilas_valid(self):
        groundtruth_uri = "https://domain.com/file1.jpeg"
        body = {
            groundtruth_uri: [
                [
                    {
                        "entity_name": 0,
                        "entity_type": "gate",
                        "entity_coords": [275, 184, 454, 183, 453, 366, 266, 367],
                    }
                ]
            ]
        }

        self.register_http_response(
            groundtruth_uri,
            method=httpretty.HEAD,
            headers={"Content-Type": "image/jpeg"},
        )
        self.validate_groundtruth_response("image_label_area_select", body)

    def test_groundtruth_uri_ilas_invalid_key(self):
        body = {
            "not_uri": [
                [
                    {
                        "entity_name": 0,
                        "entity_type": "gate",
                        "entity_coords": [275, 184, 454, 183, 453, 366, 266, 367],
                    }
                ]
            ]
        }

        with self.assertRaises(ValidationError):
            self.validate_groundtruth_response("image_label_area_select", body)

    def test_groundtruth_uri_ilas_invalid_value(self):
        body = {"https://domain.com/file1.jpeg": [[True]]}

        with self.assertRaises(ValidationError):
            self.validate_groundtruth_response("image_label_area_select", body)

    def test_taskdata_empty(self):
        """should raise if taskdata_uri contains no entries"""
        uri = "https://uri.com"
        manifest = {"taskdata_uri": uri}
        body = []

        self.register_http_response(uri, manifest, body)

        with self.assertRaises(ValidationError):
            basemodels.validate_manifest_uris(manifest)

    def test_taskdata_invalid_format(self):
        """should raise if taskdata_uri contains object instead of array"""
        uri = "https://uri.com"
        manifest = {"taskdata_uri": uri}
        body = {"key": [1, 2, 3]}

        self.register_http_response(uri, manifest, body)

        with self.assertRaises(ValidationError):
            basemodels.validate_manifest_uris(manifest)

    def test_taskdata_uri_valid(self):
        uri = "https://uri.com"
        manifest = {"taskdata_uri": uri}
        body = [
            {
                "task_key": "407fdd93-687a-46bb-b578-89eb96b4109d",
                "datapoint_uri": "https://domain.com/file1.jpg",
                "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa",
            },
            {
                "task_key": "20bd4f3e-4518-4602-b67a-1d8dfabcce0c",
                "datapoint_uri": "https://domain.com/file2.jpg",
                "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa",
            },
        ]

        self.register_http_response(uri, manifest, body)

        basemodels.validate_manifest_uris(manifest)

    def test_taskdata_uri_invalid(self):
        uri = "https://uri.com"
        manifest = {"taskdata_uri": uri}
        body = [{"task_key": "not_uuid", "datapoint_uri": "not_uri"}]

        self.register_http_response(uri, manifest, body)

        with self.assertRaises(ValidationError):
            basemodels.validate_manifest_uris(manifest)

    def test_groundtruth_and_taskdata_valid(self):
        taskdata_uri = "https://td.com"
        groundtruth_uri = "https://gt.com"
        manifest = {
            "taskdata_uri": taskdata_uri,
            "groundtruth_uri": groundtruth_uri,
            "request_type": "image_label_binary",
        }
        datapoint_uri = "https://domain.com/file1.jpg"
        taskdata = [
            {
                "task_key": "407fdd93-687a-46bb-b578-89eb96b4109d",
                "datapoint_uri": datapoint_uri,
                "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa",
            },
            {
                "task_key": "20bd4f3e-4518-4602-b67a-1d8dfabcce0c",
                "datapoint_uri": "https://domain.com/file2.jpg",
                "datapoint_hash": "f4acbe8562907183a484498ba901bfe5c5503aaa",
            },
        ]
        groundtruth_image_uri = "https://domain.com/123/file1.jpeg"
        groundtruth = {
            groundtruth_image_uri: ["false", "false", "false"],
            "https://domain.com/456/file2.jpeg": ["false", "true", "false"],
        }

        self.register_http_response(
            datapoint_uri, method=httpretty.HEAD, headers={"Content-Type": "image/jpeg"}
        )
        self.register_http_response(
            groundtruth_image_uri,
            method=httpretty.HEAD,
            headers={"Content-Type": "image/jpeg"},
        )
        self.register_http_response(taskdata_uri, manifest, taskdata)
        self.register_http_response(groundtruth_uri, manifest, groundtruth)

        basemodels.validate_manifest_uris(manifest)

    def test_mitl_in_internal_config(self):
        """Test that mitl config can be part of the internal configuration"""
        model = a_manifest().to_primitive()
        mitl_config = {
            "n_gt": 200,
            "min_tasks_in_job": 1000,
            "n_gt_sample_min": 1,
            "n_gt_sample_max": 3,
            "task_max_repeats": 25,
            "max_tasks_in_job": 36000,
            "model_id": "ResNext50_32x4d",
            "task_selection_id": "MinMargin",
            "requester_min_repeats": 12,
            "requester_max_repeats": 25,
            "stop_n_active": 1000,
            "requester_accuracy_target": 0.8,
            "nested_config": {"value_a": 1, "value_b": 2},
        }

        model["internal_config"]["mitl"] = mitl_config
        manifest = create_manifest(model)
        validate_func(manifest)()
        self.assertTrue(True)


class TaskEntryTest(unittest.TestCase):
    def test_valid_entry_is_true(self):
        taskdata = deepcopy(TASK)
        TaskDataEntry(**taskdata)

        taskdata.get("metadata")["key_1"] = 1.1
        TaskDataEntry(**taskdata)

        taskdata.get("metadata")["key_1"] = None
        TaskDataEntry(**taskdata)

        taskdata.get("metadata")["key_1"] = ""
        TaskDataEntry(**taskdata)

        with self.assertRaises(ValidationError):
            taskdata.get("metadata")["key_1"] += 1024 * "a"
            TaskDataEntry(**taskdata)

        taskdata.pop("metadata")
        TaskDataEntry(**taskdata)

        taskdata["datapoint_text"] = {"en": "Question to test with"}
        TaskDataEntry(**taskdata)

        taskdata["datapoint_uri"] = "https://domain.com/file1.jpg"
        TaskDataEntry(**taskdata)



JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION = [
    "image_label_binary",
    "image_label_multiple_choice",
    "text_free_entry",
    "image_label_area_adjust",
    "image_label_area_select",
    "image_label_single_polygon",
    "image_label_multiple_polygons",
    "image_label_semantic_segmentation_one_option",
    "image_label_semantic_segmentation_multiple_options",
    "image_label_text",
]

SUPPORTED_CONTENT_TYPES = [
    "image/jpeg",
    "image/jpg",
    "image/png",
]



import enum
import typing
import pydantic


class Pipeline(str, enum.Enum):
    FaceBlurPipeline = "FaceBlurPipeline"
    OCRThinFilterPipeline = "OCRThinFilterPipeline"
    UploadPipeline = "UploadPipeline"


class Preprocess(pydantic.BaseModel):
    pipeline: Pipeline
    config: typing.Optional[dict]

    def to_dict(self):
        p = {"pipeline": self.pipeline.value}
        if self.config is not None:
            p["config"] = self.config
        return p



import setuptools

setuptools.setup(
    name="human-protocol-basemodels",
    author="HUMAN Protocol",
    description="Common data models shared by various components of the Human Protocol stack",
    url="https://github.com/humanprotocol/human-protocol/packages/sdk/python/human-protocol-basemodels",
    include_package_data=True,
    zip_safe=True,
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    packages=setuptools.find_packages(),
    install_requires=["requests>=2", "typing-extensions", "pydantic>=1.6"],
)



"""
The VIA Format defines a standard schema for data files containing the results
of annotations produced for each datapoint of a job, i.e. the tasks defined in
a job manifest. Each file takes the general form of:

{
    "datapoints": [{
        "task_uri": "https://mydomain.com/image.jpg",
        "metadata": {
            "filename": "image.jpg"
        },
        "class_attributes": {
            "0": {
                # This nested class_attributes field will soon be deprecated
                "class_attributes": {
                    "dog": False,
                    "cat": False
                }
            }
        },
        "regions": [{
            "region_attributes": {
                "region_key": "region_value"
            },
            "shape_attributes": {
                "coords": [x1, y1, x2, y2, x3, y3, x4, y4],
                "name": "shape_type"
            }
        }],
    }]
}


"""
from typing import Dict, List, Any, Optional, Union
from pydantic import BaseModel, HttpUrl


class RegionAttributesBaseModel(BaseModel):
    region_attributes: Dict[str, Any]
    shape_attributes: Dict[str, Any]


class ClassAttributeBaseModel(BaseModel):
    # inner class attributes field not required, here for legacy purposes
    class_attributes: Optional[Dict[str, Any]]
    regions: Optional[List[RegionAttributesBaseModel]]


class DatapointBaseModel(BaseModel):
    task_uri: HttpUrl
    metadata: Optional[Dict[str, Any]]
    class_attributes: Optional[Union[Any, Dict[str, ClassAttributeBaseModel]]]


class ViaDataManifest(BaseModel):
    """Main entrypoint to define the VIA Data Format"""

    datapoints: List[DatapointBaseModel]
    version: int = 1
